// Copyright (c) 2018  GeometryFactory Sarl (France).
// All rights reserved.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is part of CGAL (www.cgal.org)
//
// $URL: https://github.com/CGAL/cgal/blob/v6.1.1/Stream_support/include/CGAL/IO/OI.h $
// $Id: include/CGAL/IO/OI.h 08b27d3db14 $
// SPDX-License-Identifier: LGPL-3.0-or-later OR LicenseRef-Commercial
//
// Author(s)     : Lutz Kettner

#ifndef CGAL_IO_OI_H
#define CGAL_IO_OI_H

#include <CGAL/IO/OI/File_writer_inventor.h>
#include <CGAL/IO/OI/Inventor_ostream.h>

#endif // CGAL_IO_OI_H
