// Copyright (c) 2009 INRIA Sophia-Antipolis (France).
// All rights reserved.
//
// This file is part of CGAL (www.cgal.org).
//
// $URL: https://github.com/CGAL/cgal/blob/v6.1.1/Mesh_3/include/CGAL/optimize_mesh_3.h $
// $Id: include/CGAL/optimize_mesh_3.h 08b27d3db14 $
// SPDX-License-Identifier: GPL-3.0-or-later OR LicenseRef-Commercial
//
//
// Author(s)     : Stephane Tayeb
//
//******************************************************************************
// File Description : Includes all optimization headers
//******************************************************************************

#ifndef CGAL_OPTIMIZE_MESH_3_H
#define CGAL_OPTIMIZE_MESH_3_H

#include <CGAL/license/Mesh_3.h>

#include <CGAL/perturb_mesh_3.h>
#include <CGAL/exude_mesh_3.h>
#include <CGAL/odt_optimize_mesh_3.h>
#include <CGAL/lloyd_optimize_mesh_3.h>

#endif // CGAL_OPTIMIZE_MESH_3_H
