// Copyright (c) 2024  GeometryFactory Sarl (France).
// All rights reserved.
//
// This file is part of CGAL (www.cgal.org).
//
// $URL: https://github.com/CGAL/cgal/blob/v6.1.1/Constrained_triangulation_3/include/CGAL/Constrained_triangulation_3_types.h $
// $Id: include/CGAL/Constrained_triangulation_3_types.h 08b27d3db14 $
// SPDX-License-Identifier: GPL-3.0-or-later OR LicenseRef-Commercial
//
// Author(s)     : Laurent Rineau

#ifndef CGAL_CT_3_TYPES_H
#define CGAL_CT_3_TYPES_H

#include <CGAL/license/Constrained_triangulation_3.h>

#include <CGAL/config.h>

namespace CGAL {

/**
 * @ingroup PkgConstrainedTriangulation3Classes
 * \brief Signed integral type to store the index of constraints.
 */
using CDT_3_signed_index = int; // must be signed

} // namespace CGAL

#endif // CGAL_CT_3_TYPES_H
