// Copyright (c) 2018 GeometryFactory (France). All rights reserved.
//
// This file is part of CGAL (www.cgal.org)
//
// $URL: https://github.com/CGAL/cgal/blob/v6.1.1/Installation/include/CGAL/enable_warnings.h $
// $Id: include/CGAL/enable_warnings.h 08b27d3db14 $
// SPDX-License-Identifier: LGPL-3.0-or-later OR LicenseRef-Commercial
//
// Author: Andreas Fabri

#include <boost/config.hpp>

#if defined(BOOST_MSVC)
#  pragma warning(pop)
#endif
