// Copyright (c) 2006-2007  INRIA Sophia-Antipolis (France).
// All rights reserved.
//
// This file is part of CGAL (www.cgal.org).
//
// $URL: https://github.com/CGAL/cgal/blob/v6.1.1/Surface_mesher/include/CGAL/make_piecewise_smooth_surface_mesh.h $
// $Id: include/CGAL/make_piecewise_smooth_surface_mesh.h 08b27d3db14 $
// SPDX-License-Identifier: GPL-3.0-or-later OR LicenseRef-Commercial
//
// Author(s)     : Laurent Rineau

#ifndef CGAL_MAKE_PIECEWISE_SMOOTH_SURFACE_MESH_H
#define CGAL_MAKE_PIECEWISE_SMOOTH_SURFACE_MESH_H

#include <CGAL/license/Surface_mesher.h>

#define CGAL_DEPRECATED_HEADER "<CGAL/make_piecewise_smooth_surface_mesh.h>"
#define CGAL_DEPRECATED_MESSAGE_DETAILS \
  "The 3D Mesh Generation package (see https://doc.cgal.org/latest/Mesh_3/) should be used instead."
#include <CGAL/Installation/internal/deprecation_warning.h>

#include <CGAL/make_surface_mesh.h>

#endif //CGAL_MAKE_PIECEWISE_SMOOTH_SURFACE_MESH_H
