// Copyright (c) 2017
// INRIA Saclay-Ile de France (France),
//
// This file is part of CGAL (www.cgal.org)
//
// $URL: https://github.com/CGAL/cgal/blob/v6.1.1/Number_types/include/CGAL/boost_mp.h $
// $Id: include/CGAL/boost_mp.h 08b27d3db14 $
// SPDX-License-Identifier: LGPL-3.0-or-later OR LicenseRef-Commercial
//
// Author(s)     : Marc Glisse

#ifndef CGAL_BOOST_MP_H
#define CGAL_BOOST_MP_H

#include <CGAL/boost_mp_type.h>
#include <CGAL/BOOST_MP_arithmetic_kernel.h>
#include <CGAL/cpp_float.h>

#endif // CGAL_BOOST_MP_H
