// Copyright (C) 2014-2018  GeometryFactory Sarl
//
// This file is part of CGAL (www.cgal.org)
//
// $URL: https://github.com/CGAL/cgal/blob/v6.1.1/Installation/include/CGAL/Surface_mesh/Surface_mesh_fwd.h $
// $Id: include/CGAL/Surface_mesh/Surface_mesh_fwd.h 08b27d3db14 $
// SPDX-License-Identifier: LGPL-3.0-or-later OR LicenseRef-Commercial
//

#ifndef CGAL_SURFACE_MESH_FWD_H
#define CGAL_SURFACE_MESH_FWD_H

/// \file Surface_mesh_fwd.h
/// Forward declarations of the Surface_mesh package.

#ifndef DOXYGEN_RUNNING
namespace CGAL {

// fwdS for the public interface
template<typename P>
class Surface_mesh;

} // CGAL
#endif

#endif /* CGAL_SURFACE_MESH_FWD_H */


